#!/usr/bin/env python3

##########################################################################################
# pihole special agent
#
# Author: Florian Dille
# Date: 12/2021
#
# Note:
#   this is the "check" with the parse function which processes the section data
#   filename: doesn't matter
#   location: <site>/local/lib/check_mk/base/plugins/agent_based/myCheckName.py
#
##########################################################################################


########################################################################################## 
##### imports

from .agent_based_api.v1 import register, Service, State, Result, check_levels, render, Metric
from time import time, mktime
from datetime import datetime

from pprint import pprint
import json


##########################################################################################    
##### parse function
#### modify (if needed) data from agent output and parse

def parse__pihole(string_table):
    data = {}
    for line, in string_table:
        str2obj = json.loads(line)
        data[str2obj["pretty"]] = str2obj
    return data


##########################################################################################
##### registering of the agent section
#### define output section to use with parse function

register.agent_section(
    name = "pihole",
    parse_function = parse__pihole
)

 
##########################################################################################
##### discovery function
#### discovers the services  

def discover__pihole(section):
    for item in section:
        yield Service(item=item)


##########################################################################################
##### check function 
#### processes everything per item

def check__pihole(item, params, section):

    data = section[item]

    if data.get("__ERROR__"):
        yield Result(state = State.UNKNOWN, summary = f"Error Message: {data['__ERROR__']}")

    else:

        #---------------------------------------------------------------------------------
        if item == "Gravity Last Update":
            p_gravity = params.get("gravity",{})

            dateA = datetime.now()
            dateB = datetime.fromtimestamp(data["absolute"])
            dateDiff = dateA - dateB

            lvl_warn = p_gravity["age"] if p_gravity.get("state",1) == 1 and p_gravity.get("age") else None
            lvl_crit = p_gravity["age"] if p_gravity.get("state",1) == 2 and p_gravity.get("age") else None

            yield from check_levels(
                label = "Age",
                value = dateDiff.total_seconds(),
                levels_upper = (lvl_warn,lvl_crit),
                metric_name = data["check_source_query"],
                render_func = render.timespan,
            )

            yield Result(state = State.OK, summary = f"Last: {render.datetime(data['absolute'])}")

        #---------------------------------------------------------------------------------
        if item == "State":
            p_state = params.get("state",{})
            thisState = 0
            thisComment = ""
            if p_state:
                if p_state.get("dst",None) is not None:
                    state_dst = "enabled" if p_state.get("dst") == True else "disabled"
                    if data['current'] != state_dst:
                        thisState = p_state.get("state",0)
                        thisComment = f", desired state: {state_dst}"
            yield Result(state = State(thisState), summary = f"{data['current']}{thisComment}")

        #---------------------------------------------------------------------------------
        if item == "DB File Size":
            p_dbfilesize = params.get("dbfilesize",{})
      
            upper_lvl_warn = (p_dbfilesize["levels_upper"][0]) if p_dbfilesize.get("levels_upper") else None
            upper_lvl_crit = (p_dbfilesize["levels_upper"][1]) if p_dbfilesize.get("levels_upper") else None

            lower_lvl_warn = (p_dbfilesize["levels_lower"][0]) if p_dbfilesize.get("levels_lower") else None
            lower_lvl_crit = (p_dbfilesize["levels_lower"][1]) if p_dbfilesize.get("levels_lower") else None

            yield from check_levels(
                label = "/etc/pihole/pihole-FTL.db",
                value = data["filesize"],
                levels_upper = (upper_lvl_warn,upper_lvl_crit),
                levels_lower = (lower_lvl_warn,lower_lvl_crit),
                metric_name = data["check_source_query"],
                render_func = render.bytes,
            )

        #---------------------------------------------------------------------------------
        if item == "Status Messages":
            p_messages = params.get("messages",{})
            msg_count = len(data.get("messages",""))

            if msg_count > 0:
                i = 1
                ## define some empty vars for later use...
                l_types = []
                l_messages = [] 
                l_messages_types_ok = []
                l_messages_types_warn = []
                l_messages_types_crit = []
                summary_appendix = []
                #state = p_messages["default_state"]
                type_state_ok = 0
                type_state_warn = 0
                type_state_crit = 0

                ## for every message... in this case to fill up some lists before the real iteration
                for message in data["messages"]:
                    ## add loop elements to certain lists...
                    l_types.append(message["type"])
                    l_messages.append(message["message"])
                    l_messages_types_ok.append(message["type"]) if message["type"] in p_messages.get("msg_ok",[]) else ""
                    l_messages_types_warn.append(message["type"]) if message["type"] in p_messages.get("msg_warn",[]) else ""
                    l_messages_types_crit.append(message["type"]) if message["type"] in p_messages.get("msg_crit",[]) else ""

                ## get more detailed output for the diffent type to state definitions via rule
                if len(l_messages_types_ok) > 0:
                    type_state_ok = 0
                    summary_appendix.append(f"ok message type count: {len(l_messages_types_ok)}")
                if len(l_messages_types_warn) > 0:
                    type_state_warn = 1
                    summary_appendix.append(f"warn message type count: {len(l_messages_types_warn)}")
                if len(l_messages_types_crit) > 0:
                    type_state_crit = 2
                    summary_appendix.append(f"crit message type count: {len(l_messages_types_crit)}")

                details = []
                if not p_messages.get("nohtml"):
                    ## HTML Output header
                    details.append("<table style='border-collapse:collapse'>")
                    details.append("<style type='text/css' scoped> .pihole_td { border:1px solid #888; padding:5px; padding-right:10px; } </style>")
                    details.append("<tr><td class='pihole_td' style='text-align:right; padding:5px;'>#</td><td class='pihole_td'>Time</td><td class='pihole_td'>Type</td><td class='pihole_td'>Message</td><td class='pihole_td'>Message Type Category</td></tr>")
                else:
                    ## get the max length for the certain columns (for non html output)
                    # TODO: is there a way to change the font for non html output for better alignment of the rows to columns
                    padding = {}
                    padding["time"] = 25
                    padding["type"] = len(max(l_types, key=len))+5
                    padding["message"] = len(max(l_messages, key=len))+5
                    ## non html output header
                    details.append(f"\n| {'#'.ljust(5)}| {'Time'.ljust(padding['time'])}| {'Type'.ljust(padding['type'])}| {'Message'.ljust(padding['message'])}| {'Message Type Category'.ljust(22)}".replace(" ","_"))

                ## for every message... now the concrete message handling
                for message in data["messages"]:
                    
                    ## additional field for the table view
                    if message["type"] in l_messages_types_crit:
                        thisMsgTypeState = "CRIT"
                    elif message["type"] in l_messages_types_warn:
                        thisMsgTypeState = "WARN"
                    elif message["type"] in l_messages_types_ok:
                        thisMsgTypeState = "OK"
                    else:
                        thisMsgTypeState = "undef"

                    ## HTML or nonHTML for every message
                    if not p_messages.get("nohtml"):
                        details.append(f"<tr><td class='pihole_td' style='text-align:right; padding:5px;'>{i}.</td><td class='pihole_td'>{render.datetime(message['timestamp'])}</td><td class='pihole_td'>{message['type']}</td><td class='pihole_td'>{message['message']}</td><td class='pihole_td'>{thisMsgTypeState}</td></tr>")
                    else:
                        details.append(f"\n| {str(i).ljust(5)}| {(render.datetime(message['timestamp'])).ljust(padding['time'])}| {message['type'].ljust(padding['type'])}| {message['message'].ljust(padding['message'])}| {thisMsgTypeState.ljust(22)}".replace(" ","_"))
                    if i == 50:
                        ## limit output to max 50 lines
                        # TODO: perhaps as a parameter?!
                        if not p_messages.get("nohtml"):
                            details.append(f"<tr><td colspan='4' class='pihole_td'>output limited to 50 there are {msg_count-i} more ...</td></tr>")
                        else:
                            details.append(f"output limited to 50 there are {msg_count-i} more ...")
                        break
                    i = i+1
                ## close table output HTML or nonHTML
                if not p_messages.get("nohtml"):
                    details.append("</table>")
                else:
                    details.append("\n")

                ## generate output(s)...
                yield Result(state = State.OK, notice = f"there are pending status messages ({', '.join(summary_appendix)})", details = f"{''.join(details)}")
                # TODO: is there a better way?
                yield Result(state = State(type_state_ok), notice = f"ok message type count: {len(l_messages_types_ok)}",)
                yield Result(state = State(type_state_warn), notice = f"warn message type count: {len(l_messages_types_warn)}",)
                yield Result(state = State(type_state_crit), notice = f"crit message type count: {len(l_messages_types_crit)}",)
            else:
                yield Result(state = State.OK, summary = "no pending messages")

            ## now the part for the total message count
            lvl_warn = int(p_messages["levels_upper"][0]) if p_messages.get("levels_upper") else None
            lvl_crit = int(p_messages["levels_upper"][1]) if p_messages.get("levels_upper") else None

            yield from check_levels(
                label = "total message count",
                value = msg_count,
                levels_upper = (lvl_warn,lvl_crit),
                metric_name = data["check_source_query"],
                render_func = lambda v: str(v),
            )
            
        #---------------------------------------------------------------------------------
        if item == "Summary":

            yield from check_levels(
                label = "Blocked Domains",
                value = data["domains_being_blocked"],
                metric_name = "domains_being_blocked",
                render_func = lambda v: str(v),
            )

            yield from check_levels(
                label = "DNS Queries Today",
                value = data["dns_queries_today"],
                metric_name = "dns_queries_today",
                render_func = lambda v: str(v),
            )

            yield from check_levels(
                label = "Ads Blocked Today",
                value = data["ads_blocked_today"],
                metric_name = "ads_blocked_today",
                render_func=lambda v: str(v),
            )

            yield from check_levels(
                label = "Ads Percentage Today",
                value = data["ads_percentage_today"],
                metric_name = "ads_percentage_today",
                boundaries=(0, 100),
                render_func = render.percent,
                notice_only = True,
            )

            yield from check_levels(
                label = "Unique Domains",
                value = data["unique_domains"],
                metric_name = "unique_domains",
                render_func = lambda v: str(v),
                notice_only = True,
            )

            yield from check_levels(
                label = "Queries Forwarded",
                value = data["queries_forwarded"],
                metric_name = "queries_forwarded",
                render_func = lambda v: str(v),
                notice_only = True,
            )

            yield from check_levels(
                label = "Queries Cached",
                value = data["queries_cached"],
                metric_name = "queries_cached",
                render_func = lambda v: str(v),
                notice_only = True,
            )

            yield from check_levels(
                label = "Clients Ever Seen",
                value = data["clients_ever_seen"],
                metric_name = "clients_ever_seen",
                render_func = lambda v: str(v),
                notice_only = True,
            )

            yield from check_levels(
                label = "Unique Clients",
                value = data["unique_clients"],
                metric_name = "unique_clients",
                render_func = lambda v: str(v),
                notice_only = True,
            )

            yield from check_levels(
                label = "DNS Queries All Types",
                value = data["dns_queries_all_types"],
                metric_name = "dns_queries_all_types",
                render_func = lambda v: str(v),
                notice_only = True,
            )

            yield from check_levels(
                label = "reply NODATA",
                value = data["reply_NODATA"],
                metric_name = "reply_NODATA",
                render_func = lambda v: str(v),
                notice_only = True,
            )

            yield from check_levels(
                label = "reply NXDOMAIN",
                value = data["reply_NXDOMAIN"],
                metric_name = "reply_NXDOMAIN",
                render_func = lambda v: str(v),
                notice_only = True,
            )

            yield from check_levels(
                label = "reply CNAME",
                value = data["reply_CNAME"],
                metric_name = "reply_CNAME",
                render_func = lambda v: str(v),
                notice_only = True,
            )

            yield from check_levels(
                label = "reply IP",
                value = data["reply_IP"],
                metric_name = "reply_IP",
                render_func = lambda v: str(v),
                notice_only = True,
            )

            yield from check_levels(
                label = "Privacy Level",
                value = data["privacy_level"],
                metric_name = "privacy_level",
                render_func = lambda v: str(v),
                notice_only = True,
            )
    

##########################################################################################
##### assemble everything

register.check_plugin(
    name = "pihole",
    sections = ["pihole"],
    service_name = "Pi-hole %s",
    discovery_function = discover__pihole,
    check_default_parameters={},
    check_function = check__pihole,
    check_ruleset_name="pihole",
)